IDI_ICON1 ICON DISCARDABLE "rendered_icons/bitcoin.ico"
IDI_ICON2 ICON DISCARDABLE "rendered_icons/bitcoin_testnet.ico"
IDI_ICON3 ICON DISCARDABLE "@CMAKE_CURRENT_SOURCE_DIR@/res/icons/bitcoin_signet.ico"
IDI_ICON4 ICON DISCARDABLE "rendered_icons/bitcoin_testnet.ico" // testnet4

#include <windows.h>             // needed for VERSIONINFO
#include <clientversion.h>  // holds the needed client version information

#define VER_PRODUCTVERSION     CLIENT_VERSION_MAJOR,CLIENT_VERSION_MINOR,CLIENT_VERSION_BUILD
#define VER_PRODUCTVERSION_STR STRINGIZE(CLIENT_VERSION_MAJOR) "." STRINGIZE(CLIENT_VERSION_MINOR) "." STRINGIZE(CLIENT_VERSION_BUILD)
#define VER_FILEVERSION        VER_PRODUCTVERSION
#define VER_FILEVERSION_STR    VER_PRODUCTVERSION_STR

VS_VERSION_INFO VERSIONINFO
FILEVERSION     VER_FILEVERSION
PRODUCTVERSION  VER_PRODUCTVERSION
FILEOS          VOS_NT_WINDOWS32
FILETYPE        VFT_APP
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4" // U.S. English - multilingual (hex)
        BEGIN
            VALUE "CompanyName",        CLIENT_NAME " project"
            VALUE "FileDescription",    CLIENT_NAME
            VALUE "FileVersion",        VER_FILEVERSION_STR
            VALUE "InternalName",       "bitcoin-qt"
            VALUE "LegalCopyright",     COPYRIGHT_STR
            VALUE "LegalTrademarks1",   "Distributed under the MIT software license, see the accompanying file COPYING or http://www.opensource.org/licenses/mit-license.php."
            VALUE "OriginalFilename",   "bitcoin-qt.exe"
            VALUE "ProductName",        CLIENT_NAME
            VALUE "ProductVersion",     VER_PRODUCTVERSION_STR
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x0, 1252 // language neutral - multilingual (decimal)
    END
END
