// Copyright (c) 2023 The Bitcoin Core developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#ifndef BITCOIN_KERNEL_MESSAGESTARTCHARS_H
#define BITCOIN_KERNEL_MESSAGESTARTCHARS_H

#include <array>
#include <cstdint>

using MessageStartChars = std::array<uint8_t, 4>;

#endif // BITCOIN_KERNEL_MESSAGESTARTCHARS_H
