#include <windows.h>             // needed for VERSIONINFO
#include "clientversion.h"       // holds the needed client version information

#define VER_PRODUCTVERSION     CLIENT_VERSION_MAJOR,CLIENT_VERSION_MINOR,CLIENT_VERSION_BUILD
#define VER_FILEVERSION        VER_PRODUCTVERSION

VS_VERSION_INFO VERSIONINFO
FILEVERSION     VER_FILEVERSION
PRODUCTVERSION  VER_PRODUCTVERSION
FILEOS          VOS_NT_WINDOWS32
FILETYPE        VFT_APP
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4" // U.S. English - multilingual (hex)
        BEGIN
            VALUE "CompanyName",        CLIENT_NAME " project"
            VALUE "FileDescription",    "bitcoin-wallet (CLI tool for " CLIENT_NAME " wallets)"
            VALUE "FileVersion",        CLIENT_VERSION_STRING
            VALUE "InternalName",       "bitcoin-wallet"
            VALUE "LegalCopyright",     COPYRIGHT_STR
            VALUE "LegalTrademarks1",   "Distributed under the MIT software license, see the accompanying file COPYING or http://www.opensource.org/licenses/mit-license.php."
            VALUE "OriginalFilename",   "bitcoin-wallet.exe"
            VALUE "ProductName",        "bitcoin-wallet"
            VALUE "ProductVersion",     CLIENT_VERSION_STRING
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x0, 1252 // language neutral - multilingual (decimal)
    END
END
