// Copyright (c) 2020-2022 The Bitcoin Core developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include <wallet/context.h>

namespace wallet {
WalletContext::WalletContext() = default;
WalletContext::~WalletContext() = default;
} // namespace wallet
